import pytest
from rs_distributions.distributions.rice import Rice
from scipy.stats import rice
import torch
import numpy as np


@pytest.mark.parametrize("test_float_broadcasting", [False, True])
@pytest.mark.parametrize("batch_shape", [3, 10])
@pytest.mark.parametrize("sample_shape", [(), (10,)])
def test_rice_execution(test_float_broadcasting, batch_shape, sample_shape):
    params = torch.ones((2, batch_shape), requires_grad=True)
    loc, scale = params
    if test_float_broadcasting:
        loc = 1.0
    q = Rice(loc, scale)
    z = q.rsample(sample_shape)
    q.mean
    q.variance
    # q.cdf(z) #<-- no cdf implementation
    q.pdf(z)
    q.log_prob(z)
    torch.autograd.grad(z.sum(), params)


def test_rice_against_scipy(
    dtype="float32", snr_cutoff=10.0, log_min=-12, log_max=2, rtol=1e-5
):
    """
    Test the following attributes of Rice against scipy.stats.rice
     - mean

    Args:
      dtype (string) : float dtype to determine the precision of the test
      snr_cutoff (float) : do not test combinations with nu/sigma values above this threshold
      log_min (int) : 10**log_min will be the minimum value tested for nu and sigma
      log_max (int) : 10**log_max will be the maximum value tested for nu and sigma
      rtol (float) : the relative tolerance for equivalency in tests
    """
    log_min, log_max = -12, 2
    nu = sigma = np.logspace(log_min, log_max, log_max - log_min + 1, dtype=dtype)
    nu, sigma = np.meshgrid(nu, sigma)
    nu, sigma = nu.flatten(), sigma.flatten()
    idx = nu / sigma < snr_cutoff
    nu, sigma = nu[idx], sigma[idx]

    q = Rice(
        torch.as_tensor(nu),
        torch.as_tensor(sigma),
    )

    mean = rice.mean(nu / sigma, scale=sigma).astype(dtype)
    result = q.mean.detach().numpy()
    assert np.allclose(mean, result, rtol=rtol)

    stddev = rice.std(nu / sigma, scale=sigma).astype(dtype)
    result = q.stddev.detach().numpy()
    assert np.allclose(stddev, result, rtol=rtol)

    z = np.linspace(
        np.maximum(mean - 3.0 * stddev, 0.0),
        mean + 3.0 * stddev,
        10,
    )

    log_prob = rice.logpdf(z, nu / sigma, scale=sigma)
    result = q.log_prob(torch.as_tensor(z)).detach().numpy()
    assert np.allclose(log_prob, result, rtol=rtol)

    pdf = rice.pdf(z, nu / sigma, scale=sigma)
    result = q.pdf(torch.as_tensor(z)).detach().numpy()
    assert np.allclose(pdf, result, rtol=rtol)



import pytest
from rs_distributions.transforms.fill_scale_tril import (
    FillScaleTriL,
    FillTriL,
    DiagTransform,
)
import torch
from torch.distributions.constraints import lower_cholesky
from torch.distributions.transforms import SoftplusTransform, ExpTransform


@pytest.mark.parametrize("batch_shape, d", [((2, 3), 6), ((1, 4, 5), 10)])
def test_forward_transform(batch_shape, d):
    transform = FillScaleTriL()
    input_shape = batch_shape + (d,)
    input_vector = torch.randn(input_shape)
    transformed_vector = transform(input_vector)

    n = int((-1 + torch.sqrt(torch.tensor(1 + 8 * d))) / 2)
    expected_output_shape = batch_shape + (n, n)
    cholesky_constraint_check = lower_cholesky.check(transformed_vector)

    assert isinstance(transformed_vector, torch.Tensor), "Output is not a torch.Tensor"
    assert transformed_vector.shape == expected_output_shape, (
        f"Expected shape {expected_output_shape}, got {transformed_vector.shape}"
    )
    assert cholesky_constraint_check.all()


@pytest.mark.parametrize("batch_shape, d", [((2, 3), 6), ((1, 4, 5), 10)])
def test_forward_equals_inverse(batch_shape, d):
    transform = FillScaleTriL()
    input_shape = batch_shape + (d,)
    input_vector = torch.randn(input_shape)
    L = transform(input_vector)
    invL = transform.inv(L)

    assert torch.allclose(input_vector, invL, atol=1e-4), (
        "Original input and the result of applying inverse transformation are not close enough"
    )


@pytest.mark.parametrize(
    "batch_shape, d, diag_transform",
    [
        ((2, 3), 6, SoftplusTransform()),
        ((1, 4, 5), 10, SoftplusTransform()),
        ((2, 3), 6, ExpTransform()),
        ((1, 4, 5), 10, ExpTransform()),
    ],
)
def test_log_abs_det_jacobian_softplus_and_exp(batch_shape, d, diag_transform):
    transform = FillScaleTriL(diag_transform=diag_transform)
    filltril = FillTriL()
    diagtransform = DiagTransform(diag_transform=diag_transform)
    input_shape = batch_shape + (d,)
    input_vector = torch.randn(input_shape, requires_grad=True)
    transformed_vector = transform(input_vector)

    # Calculate gradients log_abs_det_jacobian from FillScaleTriL
    log_abs_det_jacobian = transform.log_abs_det_jacobian(
        input_vector, transformed_vector
    )

    # Extract diagonal elements from input and transformed vectors
    tril = filltril(input_vector)
    diagonal_transformed = diagtransform(tril)

    # Calculate diagonal gradients
    diag_jacobian = diagtransform.log_abs_det_jacobian(tril, diagonal_transformed)

    # Assert diagonal gradients are approximately equal
    assert torch.allclose(diag_jacobian, log_abs_det_jacobian, atol=1e-4)



import pytest
from rs_distributions import modules as rsm
import torch


distribution_classes = rsm.DistributionModule._extract_distributions(
    rsm, base_class=rsm.DistributionModule
)

# It is common to have arguments which are equivalent and mutually exclusive
# in distribution classes.
exlusive_args = [
    ("logits", "probs"),
    ("covariance_matrix", "precision_matrix", "scale_tril"),
]

# Workarounds for distributions with additional, non-parameter arguments
special_kwargs = {
    "RelaxedBernoulli": {"temperature": torch.ones(())},
    "RelaxedOneHotCategorical": {"temperature": torch.ones(())},
    "TransformedDistribution": {
        "base_distribution": rsm.Normal(0.0, 1.0),
        "transforms": torch.distributions.AffineTransform(0.0, 1.0),
    },
    "LKJCholesky": {"dim": 3},
    "Independent": {
        "base_distribution": rsm.Normal(torch.zeros(3), torch.ones(3)),
        "reinterpreted_batch_ndims": 1,
    },
}


@pytest.mark.parametrize("distribution_class_name", distribution_classes.keys())
@pytest.mark.parametrize("serialize", [False, True])
def test_distribution_module(distribution_class_name, serialize):
    distribution_class = distribution_classes[distribution_class_name]

    assert distribution_class.__doc__ == distribution_class.distribution_class.__doc__
    assert (
        distribution_class.arg_constraints
        == distribution_class.distribution_class.arg_constraints
    )

    shape = (3, 3)
    kwargs = {}
    cons = distribution_class.arg_constraints
    for group in exlusive_args:
        matches_group = all([g in cons for g in group])
        if matches_group:
            for con in group[1:]:
                del cons[con]
    for k, v in cons.items():
        try:
            t = torch.distributions.constraint_registry.transform_to(v)
            kwargs[k] = t(torch.ones(shape))
        except NotImplementedError:
            t = torch.distributions.AffineTransform(0.0, 1.0)
            kwargs[k] = rsm.TransformedParameter(v, t)

    if distribution_class_name in special_kwargs:
        kwargs.update(special_kwargs[distribution_class_name])
    q = distribution_class(**kwargs)

    if serialize:
        from tempfile import NamedTemporaryFile

        with NamedTemporaryFile(delete=False) as f:
            torch.save(q, f)
            q = torch.load(f.name, weights_only=False)

    # Not all distributions have these attributes implemented
    try:
        q.mean
        q.variance
        q.stddev
    except NotImplementedError:
        pass

    if q.has_rsample:
        z = q.rsample()
    else:
        z = q.sample()

    ll = q.log_prob(z)

    params = list(q.parameters())
    if q.has_rsample:
        loss = -ll.sum()
        loss.backward()
        for x in params:
            assert torch.isfinite(x.grad).all()



import numpy as np
import torch
from torch import distributions as dist


class NormalIRSample(torch.autograd.Function):
    @staticmethod
    def forward(ctx, loc, scale, samples, dFdmu, dFdsig, q):
        dzdmu = -dFdmu / q
        dzdsig = -dFdsig / q
        ctx.save_for_backward(dzdmu, dzdsig)
        return samples

    @staticmethod
    def backward(ctx, grad_output):
        (
            dzdmu,
            dzdsig,
        ) = ctx.saved_tensors
        return grad_output * dzdmu, grad_output * dzdsig, None, None, None, None


class FoldedNormal(dist.Distribution):
    """
    Folded Normal distribution class

    Args:
        loc (float or Tensor): location parameter of the distribution
        scale (float or Tensor): scale parameter of the distribution (must be positive)
        validate_args (bool, optional): Whether to validate the arguments of the distribution.
        Default is None.
        var_thresh (int): Threshold for switching between the Folded Normal variance formula and the Normal variance
    """

    arg_constraints = {"loc": dist.constraints.real, "scale": dist.constraints.positive}
    support = torch.distributions.constraints.nonnegative

    def __init__(self, loc, scale, var_thresh: int = 5, validate_args=None):
        self.loc, self.scale = torch.distributions.utils.broadcast_all(loc, scale)
        batch_shape = self.loc.shape
        super().__init__(batch_shape, validate_args=validate_args)
        self._irsample = NormalIRSample().apply
        self.var_thresh = var_thresh

    def log_prob(self, value):
        """
        Compute the log-probability of the given values under the Folded Normal distribution

        Args:
            value (Tensor): The values at which to evaluate the log-probability

        Returns:
            Tensor: The log-probabilities of the given values
        """
        if self._validate_args:
            self._validate_sample(value)
        loc = self.loc
        scale = self.scale
        log_prob = torch.logaddexp(
            dist.Normal(loc, scale).log_prob(value),
            dist.Normal(-loc, scale).log_prob(value),
        )
        return log_prob

    def sample(self, sample_shape=torch.Size()):
        """
        Generate random samples from the Folded Normal distribution

        Args:
            sample_shape (torch.Size, optional): The shape of the samples to generate.
            Default is an empty shape

        Returns:
            Tensor: The generated random samples
        """
        shape = self._extended_shape(sample_shape)
        eps = torch.randn(shape, dtype=self.loc.dtype, device=self.loc.device)
        samples = torch.abs(eps * self.scale + self.loc)

        return samples

    @property
    def mean(self):
        """
        Compute the mean of the Folded Normal distribution

        Returns:
            Tensor: The mean of the distribution.
        """
        loc = self.loc
        scale = self.scale
        return scale * torch.sqrt(torch.tensor(2.0) / torch.pi) * torch.exp(
            -0.5 * (loc / scale) ** 2
        ) + loc * (1 - 2 * dist.Normal(0, 1).cdf(-loc / scale))

    @property
    def variance(self):
        """
        Compute the variance of the Folded Normal distribution

        Returns:
            Tensor: The variance of the distribution
        """
        loc = self.loc
        scale = self.scale
        mean = self.mean

        a = loc / scale
        var = torch.empty_like(loc)
        large = a > self.var_thresh
        small = ~large

        # use Normal variance when loc/scale > var_thresh
        if large.any():
            var[large] = scale[large] ** 2

        if small.any():
            var[small] = loc[small] ** 2 + scale[small] ** 2 - mean[small] ** 2

        return var

    def cdf(self, value):
        """
        Args:
            value (Tensor): The values at which to evaluate the CDF

        Returns:
            Tensor: The CDF values at the given values
        """
        if self._validate_args:
            self._validate_sample(value)
        value = torch.as_tensor(value, dtype=self.loc.dtype, device=self.loc.device)
        # return dist.Normal(loc, scale).cdf(value) - dist.Normal(-loc, scale).cdf(-value)
        return 0.5 * (
            torch.erf((value + self.loc) / (self.scale * np.sqrt(2.0)))
            + torch.erf((value - self.loc) / (self.scale * np.sqrt(2.0)))
        )

    def dcdfdmu(self, value):
        return torch.exp(
            dist.Normal(-self.loc, self.scale).log_prob(value)
        ) - torch.exp(dist.Normal(self.loc, self.scale).log_prob(value))

    def dcdfdsigma(self, value):
        A = (-(value + self.loc) / self.scale) * torch.exp(
            dist.Normal(-self.loc, self.scale).log_prob(value)
        )
        B = (-(value - self.loc) / self.scale) * torch.exp(
            dist.Normal(self.loc, self.scale).log_prob(value)
        )
        return A + B

    def pdf(self, value):
        return torch.exp(self.log_prob(value))

    def rsample(self, sample_shape=torch.Size()):
        """
        Generate differentiable random samples from the Folded Normal distribution.
        Gradients are implemented using implicit reparameterization (https://arxiv.org/abs/1805.08498).

        Args:
            sample_shape (torch.Size, optional): The shape of the samples to generate.
            Default is an empty shape

        Returns:
            Tensor: The generated random samples
        """
        samples = self.sample(sample_shape)
        # F = self.cdf(samples)
        q = self.pdf(samples)
        dFdmu = self.dcdfdmu(samples)
        dFdsigma = self.dcdfdsigma(samples)
        samples.requires_grad_(True)
        return self._irsample(self.loc, self.scale, samples, dFdmu, dFdsigma, q)



import torch
import math
from torch import distributions as dist


class RiceIRSample(torch.autograd.Function):
    @staticmethod
    def forward(ctx, nu, sigma, samples, dnu, dsigma):
        ctx.save_for_backward(dnu, dsigma)
        return samples

    @staticmethod
    def backward(ctx, grad_output):
        (
            grad_nu,
            grad_sigma,
        ) = ctx.saved_tensors
        return grad_output * grad_nu, grad_output * grad_sigma, None, None, None


class Rice(dist.Distribution):
    """
    The Rice distribution is useful for modeling acentric structure factor amplitudes in
    X-ray crystallography. It is the amplitude distribution corresponding to a bivariate
    normal in the complex plane.

    ```
    x ~ MVN([ν, 0], σI)
    y = sqrt(x[0] * x[0] + x[1] * x[1])
    ```
    The parameters ν and σ represent the location and standard deviation of an isotropic, bivariate normal.
    If x is drawn from the normal with location [ν, 0i] and covariance,
    ```
    | σ   0 |
    | 0  σi |
    ```
    the distribution of amplitudes, `y = sqrt(x * conjugate(x))`, follows a Rician distribution.

    Args:
        nu (float or Tensor): location parameter of the underlying bivariate normal
        sigma (float or Tensor): standard deviation of the underlying bivariate normal (must be positive)
        validate_args (bool, optional): Whether to validate the arguments of the distribution.
        Default is None.
    """

    arg_constraints = {
        "nu": dist.constraints.nonnegative,
        "sigma": dist.constraints.positive,
    }
    support = torch.distributions.constraints.nonnegative

    def __init__(self, nu, sigma, validate_args=None):
        self.nu, self.sigma = torch.distributions.utils.broadcast_all(nu, sigma)
        batch_shape = self.nu.shape
        super().__init__(batch_shape, validate_args=validate_args)
        self._irsample = RiceIRSample().apply

    def log_prob(self, value):
        """
        Compute the log-probability of the given values under the Rice distribution

        ```
        Rice(x | nu, sigma) = \
            x * sigma**-2 * exp(-0.5 * (x**2 + nu**2) * sigma ** -2) * I_0(x * nu * sigma **-2)
        ```

        Args:
            value (Tensor): The values at which to evaluate the log-probability

        Returns:
            Tensor: The log-probabilities of the given values
        """
        if self._validate_args:
            self._validate_sample(value)
        nu, sigma = self.nu, self.sigma
        x = value
        log_prob = (
            torch.log(x)
            - 2.0 * torch.log(sigma)
            - 0.5 * torch.square((x - nu) / sigma)
            + torch.log(torch.special.i0e(nu * x / (sigma * sigma)))
        )
        return log_prob

    def sample(self, sample_shape=torch.Size()):
        """
        Generate random samples from the Rice distribution

        Args:
            sample_shape (torch.Size, optional): The shape of the samples to generate.
            Default is an empty shape

        Returns:
            Tensor: The generated random samples
        """
        shape = self._extended_shape(sample_shape)
        nu, sigma = self.nu, self.sigma
        nu = nu.expand(shape)
        sigma = sigma.expand(shape)
        with torch.no_grad():
            A = torch.normal(nu, sigma)
            B = torch.normal(torch.zeros_like(nu), sigma)
            z = torch.sqrt(A * A + B * B)
            return z

    @property
    def mean(self):
        """
        Compute the mean of the Rice distribution

        Returns:
            Tensor: The mean of the distribution.
        """
        sigma = self.sigma
        nu = self.nu

        x = -0.5 * torch.square(nu / sigma)
        L = (1.0 - x) * torch.special.i0e(-0.5 * x) - x * torch.special.i1e(-0.5 * x)
        mean = sigma * math.sqrt(math.pi / 2.0) * L
        return mean

    @property
    def variance(self):
        """
        Compute the variance of the Rice distribution

        Returns:
            Tensor: The variance of the distribution
        """
        nu, sigma = self.nu, self.sigma
        return 2 * sigma * sigma + nu * nu - torch.square(self.mean)

    def cdf(self, value):
        """
        Args:
            value (Tensor): The values at which to evaluate the CDF

        Returns:
            Tensor: The CDF values at the given values
        """
        raise NotImplementedError("The CDF is not implemented")

    def _grad_z(self, samples):
        """
        Return the gradient of samples from this distribution

        Args:
            samples (Tensor): samples from this distribution

        Returns:
            dnu: gradient with respect to the loc parameter, nu
            dsigma: gradient with respect to the underlying normal's scale parameter, sigma
        """
        z = samples
        nu, sigma = self.nu, self.sigma
        ab = z * nu / (sigma * sigma)
        dnu = torch.special.i1e(ab) / torch.special.i0e(ab)  # == i1(ab)/i0(ab)
        dsigma = (z - nu * dnu) / sigma
        return dnu, dsigma

    def pdf(self, value):
        return torch.exp(self.log_prob(value))

    def rsample(self, sample_shape=torch.Size()):
        """
        Generate differentiable random samples from the Rice distribution.
        Gradients are implemented using implicit reparameterization (https://arxiv.org/abs/1805.08498).

        Args:
            sample_shape (torch.Size, optional): The shape of the samples to generate.
            Default is an empty shape

        Returns:
            Tensor: The generated random samples
        """
        samples = self.sample(sample_shape)
        dnu, dsigma = self._grad_z(samples)
        samples.requires_grad_(True)
        return self._irsample(self.nu, self.sigma, samples, dnu, dsigma)



import pytest
from rs_distributions.distributions.folded_normal import FoldedNormal
import torch


@pytest.mark.parametrize("test_float_broadcasting", [False, True])
@pytest.mark.parametrize("batch_shape", [3, 10])
@pytest.mark.parametrize("sample_shape", [(), (10,)])
def test_folded_normal_execution(test_float_broadcasting, batch_shape, sample_shape):
    params = torch.ones((2, batch_shape), requires_grad=True)
    loc, scale = params
    if test_float_broadcasting:
        loc = 1.0
    q = FoldedNormal(loc, scale)
    z = q.rsample(sample_shape)
    q.mean
    q.variance
    q.cdf(z)
    q.pdf(z)
    q.log_prob(z)
    torch.autograd.grad(z.sum(), params)



import torch
from rs_distributions.modules import TransformedParameter
from rs_distributions import distributions as rsd
from inspect import signature
from functools import update_wrapper


# TODO: decide whether to use "ignore" or "include" pattern here
# Distributions which are currently not supported
ignored_distributions = (
    "Uniform",  # has weird "dependent" constraints
    "Binomial",  # has_rsample == False
    "MixtureSameFamily",  # has_rsample == False
)


class DistributionModule(torch.nn.Module):
    """
    Base class for constructing learnable distributions.
    This subclass of `torch.nn.Module` acts like a `torch.distributions.Distribution`
    object with learnable `torch.nn.Parameter` attributes.
    It works by lazily constructing distributions as needed.
    Here is a simple example of distribution matching using learnable distributions with reparameterized gradients.

    ```python
    from rs_distributions import modules as rsm
    import torch

    q = rsm.FoldedNormal(10., 5.)
    p = torch.distributions.HalfNormal(1.)

    opt = torch.optim.Adam(q.parameters())

    steps = 10_000
    num_samples = 256
    for i in range(steps):
        opt.zero_grad()
        z = q.rsample((num_samples,))
        kl = (q.log_prob(z) - p.log_prob(z)).mean()
        kl.backward()
        opt.step()
    ```
    """

    distribution_class = torch.distributions.Distribution
    __doc__ = distribution_class.__doc__
    arg_constraints = distribution_class.arg_constraints

    def __init__(self, *args, **kwargs):
        super().__init__()
        sig = signature(self.distribution_class)
        bargs = sig.bind(*args, **kwargs)
        bargs.apply_defaults()
        for arg in self.distribution_class.arg_constraints:
            param = bargs.arguments.pop(arg)
            param = self._constrain_arg_if_needed(arg, param)
            setattr(self, f"_transformed_{arg}", param)
        self._extra_args = bargs.arguments

    def __repr__(self):
        rstring = super().__repr__().split("\n")[1:]
        rstring = [str(self.distribution_class) + " DistributionModule("] + rstring
        return "\n".join(rstring)

    def _distribution(self):
        kwargs = {
            k: self._realize_parameter(getattr(self, f"_transformed_{k}"))
            for k in self.distribution_class.arg_constraints
        }
        kwargs.update(self._extra_args)
        return self.distribution_class(**kwargs)

    def _constrain_arg_if_needed(self, name, value):
        if isinstance(value, TransformedParameter):
            return value
        cons = self.distribution_class.arg_constraints[name]
        if cons == torch.distributions.constraints.dependent:
            transform = torch.distributions.AffineTransform(0.0, 1.0)
        else:
            transform = torch.distributions.constraint_registry.transform_to(cons)
        return TransformedParameter(value, transform)

    @staticmethod
    def _realize_parameter(param):
        if isinstance(param, TransformedParameter):
            return param()
        return param

    def __getattr__(self, name: str):
        if name in self.distribution_class.arg_constraints or hasattr(
            self.distribution_class, name
        ):
            q = self._distribution()
            return getattr(q, name)
        return super().__getattr__(name)

    @staticmethod
    def _extract_distributions(*modules, base_class=torch.distributions.Distribution):
        """
        extract all torch.distributions.Distribution subclasses from a module(s)
        into a dict {name: cls}
        """
        d = {}
        for module in modules:
            for k in module.__all__:
                distribution_class = getattr(module, k)
                if not hasattr(distribution_class, "arg_constraints"):
                    continue
                if not hasattr(distribution_class.arg_constraints, "items"):
                    continue
                if issubclass(distribution_class, base_class):
                    d[k] = distribution_class
        return d

    def __init_subclass__(cls, /, distribution_class, **kwargs):
        super().__init_subclass__(**kwargs)
        update_wrapper(
            cls.__init__,
            distribution_class.__init__,
        )
        cls.distribution_class = distribution_class
        cls.arg_constraints = distribution_class.arg_constraints
        cls.__doc__ = distribution_class.__doc__


distributions_to_transform = DistributionModule._extract_distributions(
    torch.distributions,
    rsd,
)

for k in ignored_distributions:
    distributions_to_transform.pop(k, None)

__all__ = ["DistributionModule"]
for k, v in distributions_to_transform.items():
    # This pattern is required for pickling to work properly
    globals()[k] = type(k, (DistributionModule,), {}, distribution_class=v)
    __all__.append(k)



from .transformed_parameter import TransformedParameter  # noqa
from .distribution import DistributionModule
from .distribution import *  # noqa
from .kl import kl_divergence  # noqa
from .distribution import __all__ as all_distributions

__all__ = [
    "TransformedParameter",
    "DistributionModule",
    "kl_divergence",
]
__all__.extend(all_distributions)



import torch


class TransformedParameter(torch.nn.Module):
    """
    A `torch.nn.Module` subclass representing a constrained variabled.
    """

    def __init__(self, value, transform):
        """
        Args:
            value : Tensor
                The initial value of this learnable parameter
            transform : torch.distributions.Transform
                A transform instance which is applied to the underlying, unconstrained value
        """
        super().__init__()
        value = torch.as_tensor(value)  # support floats
        if isinstance(value, torch.nn.Parameter):
            self._value = value
            value.data = transform.inv(value)
        else:
            self._value = torch.nn.Parameter(transform.inv(value))
        self.transform = transform

    def forward(self):
        return self.transform(self._value)



from functools import wraps
from rs_distributions import modules as rsm
import torch


@wraps(torch.distributions.kl.kl_divergence)
def kl_divergence(p, q):
    if isinstance(p, rsm.DistributionModule):
        p = p._distribution()
    if isinstance(q, rsm.DistributionModule):
        q = q._distribution()
    return torch.distributions.kl.kl_divergence(p, q)



import pytest
from rs_distributions.modules import TransformedParameter
import torch


@pytest.mark.parametrize("shape", [(), 10])
def test_transformed_parameter(shape):
    value = 10.0
    eps = 1e-6
    transform = torch.distributions.ComposeTransform(
        [
            torch.distributions.AffineTransform(eps, 1.0),
            torch.distributions.ExpTransform(),
        ]
    )
    variable = TransformedParameter(value, transform)
    assert variable() == value

    params = list(variable.parameters())
    assert len(params) == 1

    loss = variable().square().sum()
    loss.backward()
    assert params[0].grad.isfinite().all()
    assert (params[0] != 0).all()



import torch
from torch.distributions import Transform, ComposeTransform, constraints
from torch.distributions.transforms import SoftplusTransform, AffineTransform
from torch.distributions.utils import vec_to_tril_matrix, tril_matrix_to_vec


class FillTriL(Transform):
    """
    Transform for converting a real-valued vector into a lower triangular matrix
    """

    def __init__(self):
        super().__init__()

    @property
    def domain(self):
        return constraints.real_vector

    @property
    def codomain(self):
        return constraints.lower_triangular

    @property
    def bijective(self):
        return True

    def _call(self, x):
        """
        Converts real-valued vector to lower triangular matrix.

        Args:
            x (torch.Tensor): input real-valued vector
        Returns:
            torch.Tensor: Lower triangular matrix
        """

        return vec_to_tril_matrix(x)

    def _inverse(self, y):
        return tril_matrix_to_vec(y)

    def log_abs_det_jacobian(self, x, y):
        batch_shape = x.shape[:-1]
        return torch.zeros(batch_shape, dtype=x.dtype, device=x.device)


class DiagTransform(Transform):
    """
    Applies transformation to the diagonal of a square matrix
    """

    def __init__(self, diag_transform):
        super().__init__()
        self.diag_transform = diag_transform

    @property
    def domain(self):
        return self.diag_transform.domain

    @property
    def codomain(self):
        return self.diag_transform.codomain

    @property
    def bijective(self):
        return self.diag_transform.bijective

    def _call(self, x):
        """
        Args:
            x (torch.Tensor): Input matrix
        Returns
            torch.Tensor: Transformed matrix
        """
        diagonal = x.diagonal(dim1=-2, dim2=-1)
        transformed_diagonal = self.diag_transform(diagonal)
        result = x.diagonal_scatter(transformed_diagonal, dim1=-2, dim2=-1)

        return result

    def _inverse(self, y):
        diagonal = y.diagonal(dim1=-2, dim2=-1)
        result = y.diagonal_scatter(self.diag_transform.inv(diagonal), dim1=-2, dim2=-1)
        return result

    def log_abs_det_jacobian(self, x, y):
        diagonal = x.diagonal(dim1=-2, dim2=-1)
        return self.diag_transform.log_abs_det_jacobian(diagonal, y)


class FillScaleTriL(ComposeTransform):
    """
    A `ComposeTransform` that reshapes a real-valued vector into a lower triangular matrix.
    The diagonal of the matrix is transformed with `diag_transform`.
    """

    def __init__(self, diag_transform=None):
        if diag_transform is None:
            diag_transform = torch.distributions.ComposeTransform(
                (
                    SoftplusTransform(),
                    AffineTransform(1e-5, 1.0),
                )
            )
        super().__init__([FillTriL(), DiagTransform(diag_transform=diag_transform)])
        self.diag_transform = diag_transform

    @property
    def bijective(self):
        return True

    def log_abs_det_jacobian(self, x, y):
        x = FillTriL()._call(x)
        diagonal = x.diagonal(dim1=-2, dim2=-1)
        return self.diag_transform.log_abs_det_jacobian(diagonal, diagonal)

    @staticmethod
    def params_size(event_size):
        """
        Returns the number of parameters required to create an n-by-n lower triangular matrix, which is given by n*(n+1)//2

        Args:
            event_size (int): size of event
        Returns:
            int: Number of parameters needed

        """
        return event_size * (event_size + 1) // 2



from rs_distributions import modules as rsm
import torch


def test_kl_divergence():
    q = rsm.Normal(0.0, 1.0)
    p = torch.distributions.Normal(0.0, 1.0)

    assert all([param.grad is None for param in q.parameters()])
    kl = rsm.kl_divergence(q, p)
    kl.backward()
    assert all([torch.isfinite(param.grad) for param in q.parameters()])
